/* subtle: a DVD subtitle-to-.sub convertor.
   Copyright 2002 Adam Sampson <azz@gnu.org>

   subtle is free software; you can redistribute and/or modify it
   under the terms of that license as published by the Free Software
   Foundation; either version 2 of the License, or (at your option)
   any later version.
   
   subtle is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with subtle; see the file COPYING. If not, write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA, or see http://www.gnu.org/.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <ctype.h>

typedef unsigned char u8;
typedef unsigned short u16;

static const char *header = "SUBTITLE";
static const int header_len = 8;

static char *opt_input = NULL;
static char *opt_picture = NULL;
static int opt_colour = 0;
static int opt_detect = 0;
static int opt_italic = 0;
static double opt_offset = 0.0;
static int opt_deftime = 300;

void die(const char *msg) {
	fprintf(stderr, "%s\n", msg);
	exit(20);
}

unsigned int rod_int(FILE *in) {
	unsigned int i;
	if (fread(&i, sizeof(i), 1, in) != 1)
		die("failed reading int");
	return i;
}

double rod_double(FILE *in) {
	double d;
	if (fread(&d, sizeof(d), 1, in) != 1)
		die("failed reading double");
	return d;
}

static char *out_text, *out_ltext = NULL;
static double out_from, out_to, out_lfrom, out_lto;

void pseudo_timecode(double time) {
	/* hh:mm:ss.cc */
	long t = time * 100;
	int h, m, s;

	h = t / 360000;
	t -= h * 360000;
	m = t / 6000;
	t -= m * 6000;
	s = t / 100;
	t -= s * 100;

	printf("%02d:%02d:%02d.%02ld", h, m, s, t);
}

void print_out() {
	if (out_ltext != NULL) {
		pseudo_timecode(out_lfrom);
		printf(",");
		if (out_lto > out_from) 
			out_lto = out_from;
		pseudo_timecode(out_lto);
		printf("\n%s\n\n", out_ltext);
		free(out_ltext);
	}
	out_ltext = out_text;
	out_lfrom = out_from;
	out_lto = out_to;
}

void process_sub(double pts, const u8 *data, u16 data_size) {
	u16 coffset = data_size;
	FILE *f;
#define SIZE 1024
	char s[SIZE];
	int x = 0, y = 0, i, fg, fgc;
	int start[2];
	int offset[2] = {0, 0};
	int counts[4] = {0, 0, 0, 0};
	static int display_on = 0;
	static u8 palette[4], alpha[4];
	static int x0, x1, y0, y1, w, h;
	u8 *picture, *p;
	int time, next;
	int extra_width;
	u8 *extra_line;

	time = (data[coffset] << 8) | data[coffset + 1];
	next = (data[coffset + 2] << 8) | data[coffset + 3];
	coffset += 4;

	/* Parse the control information */
	while (data[coffset] != 0xff) {
		switch (data[coffset++]) {
		case 0x00:
			/* Menu ID (ExtSub claims "force display", but I'm more
			   inclined to believe MPlayer) */
			display_on = 1;
			break;
		case 0x01:
			/* Start display */
			display_on = 1;
			break;
		case 0x02:
			/* Stop display */
			display_on = 0;
			break;
		case 0x03:
			/* Set palette */
			palette[0] = (data[coffset] >> 4) & 0x0f;
			palette[1] = data[coffset] & 0x0f;
			palette[2] = (data[coffset + 1] >> 4) & 0x0f;
			palette[3] = data[coffset + 1] & 0x0f;
			coffset += 2;
			break;
		case 0x04:
			/* Set alpha */
			alpha[0] = (data[coffset] >> 4) & 0x0f;
			alpha[1] = data[coffset] & 0x0f;
			alpha[2] = (data[coffset + 1] >> 4) & 0x0f;
			alpha[3] = data[coffset + 1] & 0x0f;
			coffset += 2;
			break;
		case 0x05:
			/* Set size */
			x0 = (data[coffset] << 4) | (data[coffset + 1] >> 4);
			x1 = ((data[coffset + 1] & 0x0f) << 8) | data[coffset + 2];
			y0 = (data[coffset + 3] << 4) | (data[coffset + 4] >> 4);
			y1 = ((data[coffset + 4] & 0x0f) << 8) | data[coffset + 5];
			/* Different from ExtSub: they add 1 to each of these */
			w = (x1 - x0);
			h = (y1 - y0);
			coffset += 6;
			break;
		case 0x06:
			/* Set lines */
			start[0] = (data[coffset] << 8) + data[coffset + 1];
			start[1] = (data[coffset + 2] << 8) + data[coffset + 3];
			coffset += 4;
			break;
		default:
			printf("[msg] unknown control sequence: %02x\n", data[coffset - 1]);
		}
	}
	
	/* ExtSub loops here rerunning the parse loop above until
	   next == coffset. time always seems to be 0, so we insert this kludge
	   from ExtSub. */
	if (time == 0)
		time = opt_deftime;

	p = picture = malloc(w * h);
	if (!picture)
		die("out of memory");

	while (y < h) {
		int parity = y & 1;
		int chunk, len, colour;
		const u8 *ds = &data[start[parity]];

#define READ_NIBBLE \
	do { \
		if (offset[parity] & 1) \
			chunk |= ds[offset[parity] / 2] & 0x0f; \
		else \
			chunk |= (ds[offset[parity] / 2] >> 4) & 0x0f; \
		offset[parity]++; \
	} while (0);

		chunk = 0;
		READ_NIBBLE;

		if (chunk < 0x04) {
			chunk <<= 4;
			READ_NIBBLE;
			if (chunk < 0x10) {
				chunk <<= 4;
				READ_NIBBLE;
				if (chunk < 0x40) {
					chunk <<= 4;
					READ_NIBBLE;
				}
			}
		}

		colour = chunk & 3;
		if (chunk < 4)
			len = w - x;
		else
			len = chunk >> 2;
		
		for (i = 0; i < len; i++) {
			*p++ = colour;
			counts[colour]++;
		}
		
		x += len;
		if (x >= w) {
			x = 0;
			y++;
			if (offset[parity] & 1)
				offset[parity]++;
		}
	}

	if (!opt_colour) {
		if (opt_detect) {
			/* Ugly foreground detection: pick the colour that's
 			   used least (avoiding colours that aren't used at
			   all). */
			fgc = w * h;
			for (i = 0; i < 4; i++) {
				if (counts[i] != 0 && counts[i] < fgc) {
					fgc = counts[i];
					fg = i;
				}
			}
		}
	
		/* Make the picture into black on white for gocr's benefit. */
		p = picture;
		for (i = 0; i < (w * h); i++, p++) 
			if (opt_detect ? *p == fg : *p > 1)
				*p = 0;
			else
				*p = 1;
	}

	/* Work out the italic kludge. */
	if (opt_italic == 0) {
		extra_width = 0;
	} else {
		extra_width = h / opt_italic;
		extra_line = malloc(extra_width);
		memset(extra_line, 1, extra_width);
	}

	if (opt_picture == NULL) {
		f = popen("gocr -d 0 -s 10 -i - -o subtle_tmp", "w");
		if (!f)
			die("unable to open pipe to gocr");
	} else {
		snprintf(s, SIZE, "%s-%010.5f.pgm", opt_picture, pts);
		f = fopen(s, "w");
		if (!f)
			die("unable to open pgm output file");
	}

	fprintf(f, "P5\n%d %d 4\n", w + extra_width, h);	
	if (opt_italic != 0) {
		for (i = 0; i < h ; i++) {
			fwrite(extra_line, 1, i / opt_italic, f);
			fwrite(picture + (i * w), 1, w, f);
			fwrite(extra_line, 1, extra_width - (i / opt_italic), f);
		}
	} else {
		fwrite(picture, 1, w * h, f);
	}
	free(picture);

	if (opt_picture == NULL) {
		int tsize = 1024;

		pclose(f);

		out_from = pts - opt_offset;
		out_to = out_from + (time / 100.0);
		out_text = malloc(tsize);
		if (out_text == NULL)
			die("out of memory");
		*out_text = '\0';

		f = fopen("subtle_tmp", "r");
		while (fgets(s, SIZE, f) != NULL) {
			char *p;

			for (p = s + strlen(s) - 1; p > s && isspace(*p); p--)
				*p = '\0';
			for (p = s; *p && isspace(*p); p++)
				;
		
			if ((strlen(out_text) + strlen(p) + 5) >= tsize) {
				tsize *= 2;
				out_text = realloc(out_text, tsize);
				if (out_text == NULL)
					die("out of memory");
			}
			if (*out_text != '\0')
				strcat(out_text, "[br]");
			strcat(out_text, p);
		}

		print_out();

		fclose(f);
		unlink("subtle_tmp");
	} else {
		fclose(f);
	}

	if (opt_italic != 0)
		free(extra_line);
}

static u8 sub_data[65536];
static int sub_size = 0;
static int sub_datasize = 0;
static int sub_total = -1;
static double sub_pts;

void build_sub(double pts, const u8 *payload, int payload_length) {
	if (sub_total == -1) {
		/*
		u8 type = payload[0];
		u8 id = type & 0x1f;
		*/
		u16 total_size = (payload[1] << 8) | payload[2];
		u16 data_size = (payload[3] << 8) | payload[4];

		sub_total = total_size;
		sub_datasize = data_size;
		sub_size = 0;	
	}

	payload++;
	payload_length--;

	/* This looks wrong... should check for overflow. */
	memcpy(sub_data + sub_size, payload, payload_length);
	sub_size += payload_length;
	sub_pts = pts;

	if (sub_size < sub_total)
		return;

	/* sub_data now contains sub_datasize bytes of data, then the control
	   information. */
	process_sub(pts, sub_data, sub_datasize);

	sub_total = -1;
}

int main(int argc, char **argv) {
	FILE *f;

	while (1) {
		int c = getopt(argc, argv, "i:p:t:o:d:ceh");
		if (c == -1)
			break;
		switch (c) {
		case 'i':
			opt_input = strdup(optarg);
			break;
		case 'p':
			opt_picture = strdup(optarg);
			break;
		case 't':
			opt_italic = atoi(optarg);
			break;
		case 'o':
			opt_offset = atof(optarg);
			break;
		case 'd':
			opt_deftime = atoi(optarg);
			break;
		case 'c':
			opt_colour = 1;
			break;
		case 'e':
			opt_detect = 1;
			break;
		case 'h':
			fprintf(stderr, "Usage: subtitle_parse [-i input_file] [-p image_prefix] [-t italic_correction] [-o time_offset] [-c] [-e] [-h]\n");
			return 0;
		default:
			die("unknown option");
		}
	}

	if (opt_input == NULL) {
		f = stdin;
	} else {
		f = fopen(opt_input, "r");
		if (!f)
			die("unable to open input file");
	}

	if (opt_picture == NULL)
		printf("[INFORMATION]\n[TITLE]\n[AUTHOR]\n[SOURCE]subtle\n[PRG]\n[FILEPATH]\n[DELAY]0\n[CD TRACK]0\n[COMMENT]\n[END INFORMATION]\n[SUBTITLE]\n");

	while (1) {
		unsigned int header_length, header_version, payload_length,
			lpts, discont_ctr;
		double rpts;
		u8 *payload, buffer[header_len];

		if (fread(buffer, header_len, 1, f) != 1)
			break;
		if (strncmp(buffer, header, header_len) != 0)
			die("SUBTITLE wasn't SUBTITLE");
		header_length = rod_int(f);
		header_version = rod_int(f);
		payload_length = rod_int(f);
		lpts = rod_int(f);
		rpts = rod_double(f);
		discont_ctr = rod_int(f);
	
		payload = malloc(payload_length);
		if (!payload)
			die("out of memory");
		if (fread(payload, 1, payload_length, f) != payload_length)
			die("unable to read payload");
		build_sub(rpts, payload, payload_length);
		free(payload);
	}

	if (opt_input != NULL)
		fclose(f);

	if (opt_picture == NULL) {
		out_text = NULL;
		out_from = out_lto + 1;
		print_out();
	}

	return 0;
}

